package core

// version related consts
const (
	Name    = "opensnitch-daemon"
	Version = "1.7.2"
	Author  = "Simone 'evilsocket' Margaritelli"
	Website = "https://github.com/evilsocket/opensnitch"
)
